require 'pathname'
require 'fileutils'

module UU
  module OS
    module Env

      # Provides operations for the current execution environment.
      class Environment

        # Flag if environment configuration was initialized.
        # @private
        @@CONFIG_INITIALIZED = false
        
        # Relative path to configuration file
        # @private
        CFG_FILE_PATH = 'config/uu-client.properties'

        # Returns directory which should (by default) contain all application configuration files.
        # It may be defined by environment variable +UU_HOME+ and defaults to '~/.uu'
        #
        # @return [String] UU Home directory
        def self.uu_home
          path = Java::JavaLang::System.getProperty('uu.home') if RUBY_PLATFORM[/java/] 
          path = ENV['UU_HOME'] if path.nil?
          if path.nil?
            result = Pathname.new(File.join(Dir.home, '.uu'))
          else
            result = Pathname.new(path)
          end
          if !@@CONFIG_INITIALIZED
            self.send(:init_config, result)
            @@CONFIG_INITIALIZED = true
          end
          result
        end

        # Loads configuration file (in property format).
        #
        # @param [String] path Path to configuration file (absolute or relative to UU Home)
        #
        # @return [Hash] Hash containing parsed configuration file
        def self.load_configuration(path)
          if !Pathname.new(path).absolute?
            path = File.join(self.uu_home, path)
          end
          if !File.exist?(path)
            raise ArgumentError.new("Configuration file #{path} does not exist.")
          end
          properties = {}
          File.open(path, 'rt') do |file|
            content = file.read
            properties = self.parse_properties(content)
          end
          properties
        end

        # Parses data in Java properties format and returns it as a hash.
        #
        # @param data [String] The data to parse.
        # @return [Hash] A hash containing the parsed properties.
        def self.parse_properties(data)
          properties = {}
          data.gsub!(/\r\n?/, "\n")
          data.each_line do |line|
            line.strip!
            if (line[0] != ?# and line[0] != ?=)
              i = line.index('=')
              if (i)
                properties[line[0..i - 1].strip] = line[i + 1..-1].strip
              else
                properties[line] = ''
              end
            end
          end
          properties
        end

        # Initializes environment by copying of default configuration files
        # included in gem to UU Home directory.
        #
        # @param [String] uu_home UU Home directory where to copy configuration files
        #
        # @private
        def self.init_config(uu_home)
          source_root = File.join(File.dirname(__FILE__), '/../../../../etc')
          target_root = uu_home
          Dir[File.join(source_root, '/**/*')].each do |file|
            if (!File.directory?file)
              target = File.join(target_root, file.split(source_root))
              if (!File.exists?target)
                FileUtils.mkpath(File.dirname(target))
                FileUtils.copy(file, target)
              end
            end
          end
        end
        private_class_method :init_config

      end

    end
  end
end
