require 'json'
require 'base64'
require 'tempfile'
require 'uu/os/cmd/command_client'
require 'uu/os/cmd/command_handler'
require 'uu/os/http/http_query'
require 'uu/os/util/url_converter'
require 'uu/os/env/environment'
require 'uu/os/env/process'
require 'uu/os/qos/qos_limit_exception'
require 'uu/os/rest/binary_value'
require 'uu/os/rest/client_exception'
require 'uu/os/rest/main_entity_exception'
require 'uu/os/rest/server_exception'
require 'uu/os/security/authentication_exception'
require 'uu/os/security/authorization_exception'
require 'uu/os/security/session'

# Load configuration from uu-client.properties and inject them
# into CommandClient to make configuration backward compatible.
config = UU::OS::Env::Environment.load_configuration('config/uu-client.properties')
UU::OS::CMD::CommandClient.load_configuration(config)

module UU
  module OS
    module REST

      # Client for invocation of REST uuAPI.
      # @deprecated To be used only for commands v1. For commands v2 use
      #   {UU::OS::CMD::CommandClient UU::OS::CMD::CommandClient}
      class RemoteClient
        include UU::OS::Util::URLConverter

        # Relative path to configuration file
        # @private
        @@CFG_FILE_PATH = 'config/uu-client.properties'

        # Default base URL
        # @private
        @@DEFAULT_BASE_URL = 'https://api.plus4u.net'

        # Name of thread parameter for storing default path prefix
        # @private
        @@PATH_PREFIX_PARAM = 'uu.os.rest.remote-client-path-prefix'

        # Default path prefix
        # @private
        @@DEFAULT_PATH_PREFIX = '/ues/wcp'

        # Server path parameter
        # @private
        @@SERVER_PATH_PARAM = 'server-path-prefix'

        # Trust server certificate parameter
        # @private
        @@TRUST_SERVER_CERT_PARAM = 'trust-server-cert'

        # Trust server hostname parameter
        # @private
        @@TRUST_SERVER_HOSTNAME_PARAM = 'trust-server-hostname'

        # Process URI header parameter
        # @private
        @@PROCESS_HEADER_PARAM = :'uu-os-process-uri'

        # Default request/response content type
        # @private
        @@DEFAULT_CONTENT_TYPE = 'application/json'

        # Multipart request content type
        # @private
        @@MULTIPART_CONTENT_TYPE = 'multipart/form-data'

        # Stream content type
        # @private
        @@STREAM_CONTENT_TYPE = 'application/octet-stream'

        # Default request/response encoding
        # @private
        @@DEFAULT_ENCODING = 'UTF-8'

        # Default request timeout
        # @private
        @@DEFAULT_TIMEOUT = 120
        # Returns default base URL (this URL is used in case no URL is found
        # in configuration file for a particular service application or system).
        #
        # @return [String] Default base URL
        # @deprecated To be used only for commands v1. For commands v2 use
        #   {UU::OS::CMD::CommandClient UU::OS::CMD::CommandClient}
        def self.default_base_url
          @@DEFAULT_BASE_URL
        end

        # Stores default base URL (this URL is used in case no URL is found
        # in configuration file for a particular service application or system).
        #
        # @param [String] base_url Default base URL
        # @deprecated To be used only for commands v1. For commands v2 use
        #   {UU::OS::CMD::CommandClient UU::OS::CMD::CommandClient}
        def self.default_base_url=(base_url)
          @@DEFAULT_BASE_URL = base_url
        end

        # Returns default path prefix to be injected between +default_base_url+
        # and PATH constant defined on UU service classes.
        #
        # @return [String] Default path prefix
        # @deprecated To be used only for commands v1. For commands v2 use
        #   {UU::OS::CMD::CommandClient UU::OS::CMD::CommandClient}
        def self.default_path_prefix
          if !Thread.current[@@PATH_PREFIX_PARAM]
            self.default_path_prefix=@@DEFAULT_PATH_PREFIX
          end
          Thread.current[@@PATH_PREFIX_PARAM]
        end

        # Stores default path prefix to be injected between +default_base_url+
        # and PATH constant defined on UU service classes.
        #
        # @param [String] path_prefix Default path prefix
        # @deprecated To be used only for commands v1. For commands v2 use
        #   {UU::OS::CMD::CommandClient UU::OS::CMD::CommandClient}
        def self.default_path_prefix=(path_prefix)
          Thread.current[@@PATH_PREFIX_PARAM] = path_prefix
        end

        # Creates new instance of REST client.
        #
        # @param [String, Class] base_url URL of UU server or Service class (URL will be read from configuration based on naming conventions)
        # @param [String] service_path Service path (may be +nil+ in case +base_url+ is service class with defined +PATH+ constant)
        # @deprecated To be used only for commands v1. For commands v2 use
        #   {UU::OS::CMD::CommandClient UU::OS::CMD::CommandClient}
        def initialize(base_url, service_path = nil)
          options = {}
          if (base_url.kind_of?Class) || (base_url.kind_of?Module)
            @service = base_url
            # Get service path
            service_path = @service::PATH if service_path.nil?
            base_url = nil
            # Load path_prefix from configuration file.
            path_prefix = get_config_param(service_path, @@SERVER_PATH_PARAM)
            path_prefix = self.class.default_path_prefix if path_prefix.nil?
          else
            path_prefix = ''
          end
          trust_hostname = get_config_param(service_path, @@TRUST_SERVER_HOSTNAME_PARAM)
          trust_cert = get_config_param(service_path, @@TRUST_SERVER_CERT_PARAM)
          if trust_hostname.nil? && trust_cert.nil?
            # Due to backward compatibility reasons, in case parameters are not defined,
            # SSL validation should be disabled (so unlike base URL and SSL version we
            # cannot leave configuration on CommandClient but need to force it).
            options[:validate_ssl] = false
          elsif trust_hostname == 'false' && trust_cert == 'false'
            options[:validate_ssl] = true
          else
            options[:validate_ssl] = false
          end
          options[:default_action] = :'/'
          options[:session_handler] = self
          options[:command_handlers] = self
          options[:main_object_uri_parameter_name] = :uesuri
          options[:parameters_parameter_name] = ''
          options[:base_url] = base_url
          options[:default_base_url] = @@DEFAULT_BASE_URL
          options[:use_legacy_config] = true
          options[:append_routing_data] = false
          @base_path = concat_path(path_prefix, service_path)
          @client = UU::OS::CMD::CommandClient.new(@base_path, nil, options)
          @client.unwrap.set_header(:accept, @@DEFAULT_CONTENT_TYPE)
          @client.unwrap.set_header(:'content-type', @@DEFAULT_CONTENT_TYPE)
          @default_query = UU::OS::HTTP::HTTPQuery.new
          self.timeout = @@DEFAULT_TIMEOUT
        end

        # Sets HTTP authorization token to use for authentication and authorization. This token
        # will be used in request header for each command invoked with this client instance.
        #
        # @param [String] token HTTP authorization token
        # @deprecated To be used only for commands v1. For commands v2 use
        #   {UU::OS::CMD::CommandClient UU::OS::CMD::CommandClient}
        def auth_token=(token)
          @auth_token = token
        end

        # Adds HTTP header. This header will be used in request for each command invoked with this client instance.
        #
        # @param [String, Symbol] name Name of header
        # @param [String] value Header value
        # @deprecated To be used only for commands v1. For commands v2 use
        #   {UU::OS::CMD::CommandClient UU::OS::CMD::CommandClient}
        def add_header(name, value)
          @client.unwrap.set_header(name, value)
        end

        # Adds HTTP query parameter. This parameter will be used in request for each command invoked with this client instance.
        #
        # @param [String, Symbol] name Name of parameter
        # @param [String] value Parameter value
        # @deprecated To be used only for commands v1. For commands v2 use
        #   {UU::OS::CMD::CommandClient UU::OS::CMD::CommandClient}
        def add_parameter(name, value)
          @default_query[name] = value
        end

        # Sets request timeout (default value is 120 seconds)
        #
        # @param [Numeric] timeout Request timeout (-1 for infinite timeout)
        # @deprecated To be used only for commands v1. For commands v2 use
        #   {UU::OS::CMD::CommandClient UU::OS::CMD::CommandClient}
        def timeout=(timeout)
          if timeout < 0
            # HTTPClient uses zero for infinite timeout
            tval = 0
          elsif timeout == 0
            # therefore zero needs to be translated to minimal timeout
            tval = 0.001
          else
            tval = timeout
          end
          @client.unwrap.connect_timeout = tval
          @client.unwrap.request_timeout = tval
        end

        # Invokes HTTP GET method without touching the incoming response.
        #
        # @param [String, Symbol] method Service method to invoke
        # @param [String, UU::OS::UESURI] uesuri Entity URI (optional)
        #
        # @return [Array<Numeric, String, Hash>] response status code, response body, response headers
        # @deprecated To be used only for commands v1. For commands v2 use
        #   {UU::OS::CMD::CommandClient UU::OS::CMD::CommandClient}
        def raw_get(method, uesuri = nil)
          params = {:invocation_method => :get, :raw_response => true}
          response = @client.invoke(method, UU::OS::UESURI.new(uesuri), params).unwrap
          body = response.body
          body = body.read if body.respond_to?(:read)
          body.force_encoding(@@DEFAULT_ENCODING)
          return [response.status, body, transform_headers(response.headers)]
        end

        # Invokes HTTP GET method without touching the incoming response.
        # Response body is handed as binary read only stream. Caller is
        # responsible for closing this stream after it is processed.
        #
        # @param [String, Symbol] method Service method to invoke
        # @param [String, UU::OS::UESURI] uesuri Entity URI (optional)
        #
        # @return [Array<Numeric, IO, Hash>] response status code, response body, response headers
        # @deprecated To be used only for commands v1. For commands v2 use
        #   {UU::OS::CMD::CommandClient UU::OS::CMD::CommandClient}
        def raw_get_binary(method, uesuri = nil)
          params = {:invocation_method => :get, :raw_response => true, :stream_response => true, :binary_response => true}
          if @client.unwrap.instance_variable_get(:@headers)[:accept][0] == @@DEFAULT_CONTENT_TYPE
            # RemoteClient allowed to reconfigure Accept header for whole instance
            # (probably never used but we need it for backward compatibility).
            params[:accept] = @@STREAM_CONTENT_TYPE
          end
          response = @client.invoke(method, UU::OS::UESURI.new(uesuri), params).unwrap
          # Some commands rely on ability to rewind stream of downloaded
          # binary data (rest-client used tempfiles so this was possible).
          # Therefore to maintain backward compatibility, we also have to
          # also create tempfiles.
          return [response.status, TempfileIO.new(response.body), transform_headers(response.headers)]
        end

        # Invokes HTTP POST method without touching the incoming response.
        #
        # @param [String, Symbol] method Service method to invoke
        # @param [String, UU::OS::UESURI] uesuri Entity URI (optional)
        # @param [String, Hash] data Data payload (optional; +Hash+ object will be sent as multipart/form-data)
        #
        # @return [Array<Numeric, String, Hash>] response status code, response body, response headers
        # @deprecated To be used only for commands v1. For commands v2 use
        #   {UU::OS::CMD::CommandClient UU::OS::CMD::CommandClient}
        def raw_post(method, uesuri = nil, data = nil)
          params = {:invocation_method => :post, :raw_response => true}
          if data.kind_of?Hash
            params[:multipart_upload] = true
          end
          params[:parameters] = data
          response = @client.invoke(method, UU::OS::UESURI.new(uesuri), params).unwrap
          body = response.body
          body = body.read if body.respond_to?(:read)
          body.force_encoding(@@DEFAULT_ENCODING)
          return [response.status, body, transform_headers(response.headers)]
        end

        # Invokes HTTP POST method without touching the incoming response.
        # Response body is handed as binary read only stream. Caller is
        # responsible for closing this stream after it is processed.
        #
        # @param [String, Symbol] method Service method to invoke
        # @param [String, UU::OS::UESURI] uesuri Entity URI (optional)
        # @param [String, Hash] data Data payload (optional; +Hash+ object will be sent as multipart/form-data)
        #
        # @return [Array<Numeric, IO, Hash>] response status code, response body, response headers
        # @deprecated To be used only for commands v1. For commands v2 use
        #   {UU::OS::CMD::CommandClient UU::OS::CMD::CommandClient}
        def raw_post_binary(method, uesuri = nil, data = nil)
          params = {:invocation_method => :post, :raw_response => true, :stream_response => true, :binary_response => true}
          if data.kind_of?Hash
            params[:multipart_upload] = true
          end
          if @client.unwrap.instance_variable_get(:@headers)[:accept][0] == @@DEFAULT_CONTENT_TYPE
            # RemoteClient allowed to reconfigure Accept header for whole instance
            # (probably never used but we need it for backward compatibility).
            params[:accept] = @@STREAM_CONTENT_TYPE
          end
          params[:parameters] = data
          response = @client.invoke(method, UU::OS::UESURI.new(uesuri), params).unwrap
          # Some commands rely on ability to rewind stream of downloaded
          # binary data (rest-client used tempfiles so this was possible).
          # Therefore to maintain backward compatibility, we also have to
          # also create tempfiles.
          return [response.status, TempfileIO.new(response.body), transform_headers(response.headers)]
        end

        # Invokes HTTP DELETE method without touching the incoming response.
        #
        # @param [String, Symbol] method Service method to invoke
        # @param [String, UU::OS::UESURI] uesuri Entity URI (optional)
        #
        # @return [Array<Numeric, String, Hash>] response status code, response body, response headers
        # @deprecated To be used only for commands v1. For commands v2 use
        #   {UU::OS::CMD::CommandClient UU::OS::CMD::CommandClient}
        def raw_delete(method, uesuri = nil)
          # Delete is not implemented in CommandClient, so we have to use HTTPClient
          http_client = @client.unwrap
          delete = http_client.prepare_delete(@base_path, method)
          delete.parameters[:uesuri] = UU::OS::UESURI.new(uesuri)
          delete.parameters.merge!(@default_query)
          delete.headers[:authorization] = @auth_token || UU::OS::Security::Session.send(:get_authn_token)
          if !UU::OS::Env::Process.active_process_uri.nil?
            delete.headers[@@PROCESS_HEADER_PARAM] = UU::OS::Env::Process.active_process_uri.to_s
          end
          response = delete.execute
          return [response.status, response.body, transform_headers(response.headers)]
        end

        # Invokes HTTP GET and checks incoming response for errors.
        #
        # @param [String, Symbol] method Service method to invoke
        # @param [String, UU::OS::UESURI] uesuri Entity URI (optional)
        #
        # @return [String] Response body (or throws exception)
        # @deprecated To be used only for commands v1. For commands v2 use
        #   {UU::OS::CMD::CommandClient UU::OS::CMD::CommandClient}
        def get(method, uesuri = nil)
          params = {:invocation_method => :get}
          return @client.invoke(method, UU::OS::UESURI.new(uesuri), params)
        end

        # Invokes HTTP GET and checks incoming response for errors.
        # Response body is handed as binary value containing response
        # body stream. Caller is responsible for closing this stream
        # after it is processed.
        #
        # @param [String, Symbol] method Service method to invoke
        # @param [String, UU::OS::UESURI] uesuri Entity URI (optional)
        #
        # @return [BinaryValue] Binary value with response body stream (or throws exception)
        # @deprecated To be used only for commands v1. For commands v2 use
        #   {UU::OS::CMD::CommandClient UU::OS::CMD::CommandClient}
        def get_binary(method, uesuri = nil)
          params = {:invocation_method => :get, :stream_response => true, :binary_response => true}
          if @client.unwrap.instance_variable_get(:@headers)[:accept][0] == @@DEFAULT_CONTENT_TYPE
            # RemoteClient allowed to reconfigure Accept header for whole instance
            # (probably never used but we need it for backward compatibility).
            params[:accept] = @@STREAM_CONTENT_TYPE
          end
          return @client.invoke(method, UU::OS::UESURI.new(uesuri), params)
        end

        # Invokes HTTP POST and checks incoming response for errors.
        #
        # @param [String, Symbol] method Service method to invoke
        # @param [String, UU::OS::UESURI] uesuri Entity URI (optional)
        # @param [String, Hash] data Data payload (optional; +Hash+ object will be sent as multipart/form-data)
        #
        # @return [String] Response body (or throws exception)
        # @deprecated To be used only for commands v1. For commands v2 use
        #   {UU::OS::CMD::CommandClient UU::OS::CMD::CommandClient}
        def post(method, uesuri = nil, data = nil)
          params = {:invocation_method => :post}
          if data.kind_of?Hash
            params[:multipart_upload] = true
          end
          params[:parameters] = data
          return @client.invoke(method, UU::OS::UESURI.new(uesuri), params)
        end

        # Invokes HTTP POST and checks incoming response for errors.
        # Response body is handed as binary value containing response
        # body stream. Caller is responsible for closing this stream
        # after it is processed.
        #
        # @param [String, Symbol] method Service method to invoke
        # @param [String, UU::OS::UESURI] uesuri Entity URI (optional)
        # @param [String, Hash] data Data payload (optional; +Hash+ object will be sent as multipart/form-data)
        #
        # @return [BinaryValue] Binary value with response body stream (or throws exception)
        # @deprecated To be used only for commands v1. For commands v2 use
        #   {UU::OS::CMD::CommandClient UU::OS::CMD::CommandClient}
        def post_binary(method, uesuri = nil, data = nil)
          params = {:invocation_method => :post, :stream_response => true, :binary_response => true}
          if data.kind_of?Hash
            params[:multipart_upload] = true
          end
          params[:parameters] = data
          if @client.unwrap.instance_variable_get(:@headers)[:accept][0] == @@DEFAULT_CONTENT_TYPE
            # RemoteClient allowed to reconfigure Accept header for whole instance
            # (probably never used but we need it for backward compatibility).
            params[:accept] = @@STREAM_CONTENT_TYPE
          end
          return @client.invoke(method, UU::OS::UESURI.new(uesuri), params)
        end

        # Invokes HTTP DELETE and checks incoming response for errors.
        #
        # @param [String, Symbol] method Service method to invoke
        # @param [String, UU::OS::UESURI] uesuri Entity URI (optional)
        #
        # @return [String] Response body (or throws exception)
        # @deprecated To be used only for commands v1. For commands v2 use
        #   {UU::OS::CMD::CommandClient UU::OS::CMD::CommandClient}
        def delete(method, uesuri = nil)
          # Delete is not implemented in CommandClient, so we have to use HTTPClient
          http_client = @client.unwrap
          delete = http_client.prepare_delete(@base_path, method)
          delete.parameters[:uesuri] = UU::OS::UESURI.new(uesuri)
          delete.parameters.merge!(@default_query)
          delete.headers[:authorization] = @auth_token || UU::OS::Security::Session.send(:get_authn_token)
          if !UU::OS::Env::Process.active_process_uri.nil?
            delete.headers[@@PROCESS_HEADER_PARAM] = UU::OS::Env::Process.active_process_uri.to_s
          end
          response = delete.execute
          return @client.wrap_result(response, :invocation_method => :delete, :main_object_uri_parameter_name => :uesuri).get
        end

        # Override inspect to hide internal structures.
        # @private
        def inspect # :nodoc:
          "#<#{self.class.name}:0x#{(self.object_id << 1).to_s(16)}>"
        end

        private

        # Loads configuration parameter for given service from default
        # configuration file.
        #
        # @param [String] service_path REST Path of service for which parameter is loaded
        # @param [String] parameter_name Name of configuration parameter
        #
        # @return [String] Parameter value, nil if parameter or default configuration file does not exist
        #
        # @private
        def get_config_param(service_path, parameter_name) # :nodoc:
          value = nil
          begin
            propt = UU::OS::Env::Environment.load_configuration(@@CFG_FILE_PATH)
            tmp = service_path.dup
            tmp = tmp[1..-1] if tmp[/^[\/]/]
            tmp.gsub!(/^ues\/platform\//, 'uu/os/')
            tmp.gsub!(/^ues\/core\//, 'uu/os/')
            tmp.gsub!(/^ues\//, 'uu/')
            parts = tmp.split('/')
            for i in 1..parts.size
              key = parts[0..-i].join('.').downcase.concat(".#{parameter_name}")
              value = propt[key]
              break if !value.nil?
            end
          rescue
            # No problem, default value will be used
          end
          value
        end

        # HTTPClient returns header values as array. This method transforms
        # headers so that they are backward compatible with value returned
        # by RemoteClient.
        # @private
        def transform_headers(http_headers) # :nodoc:
          result = {}
          http_headers.each do |key, value|
            key = key.to_s.gsub('-','_').to_sym
            result[key] = value.join(',')
          end
          return result
        end

        public

        # Implementation of {UU::OS::CMD::CMD::SessionHandler}.
        # @private
        def handle_session(session, http_headers, http_parameters) # :nodoc:
          http_headers[:authorization] = @auth_token || UU::OS::Security::Session.send(:get_authn_token)
          if !UU::OS::Env::Process.active_process_uri.nil?
            http_headers[@@PROCESS_HEADER_PARAM] = UU::OS::Env::Process.active_process_uri.to_s
          end
        end

        # Implementation of {UU::OS::CMD::CommandHandler}.
        # @private
        def handle_request(payload, http_headers, http_parameters, options) # :nodoc:
          # RemoteClient allowed to send query parameters also for POST requests
          # (probably never used but we need it for backward compatibility).
          http_parameters.merge!(@default_query)
          if options[:multipart_upload] = true && payload.kind_of?(Hash)
            # rest-client was sending multipart with nested structures as flat.
            # Therefore we have to flatten payload to maintain backward compatibility.
            flat_payload = {}
            flatten_nested_structures(payload, flat_payload)
            return flat_payload
          else
            return UU::OS::CMD::CommandHandler::NEXT
          end
        end

        # Implementation of {UU::OS::CMD::CMD::CommandHandler}.
        # @private
        def handle_response(http_response, options) # :nodoc:
          if options[:raw_response]
            result = http_response.body
            if options[:binary_response]
              result = StringIO.new(result) if !result.kind_of?(IO)
            else
              result = result.read if !result.kind_of?(String)
              result.force_encoding(@@DEFAULT_ENCODING)
            end
          else
            if options[:binary_response]
              headers = http_response.headers
              # Try to parse binary value metadata from optional header
              metadata = headers[:uu_os_binary_data_info]
              metadata &&= JSON.parse(Base64.decode64(metadata[0]), :symbolize_names => true)
              data = http_response.body
              if data.kind_of?(IO)
                # Some commands rely on ability to rewind stream of downloaded
                # binary data (rest-client used tempfiles so this was possible).
                # Therefore to maintain backward compatibility, we also have to
                # also create tempfiles.
                data = TempfileIO.new(data)
              else
                data = StringIO.new(data)
              end
              if metadata.nil? && data.size == 0
                return nil
              end
              result = UU::OS::REST::BinaryValue.new(metadata, true)
              # Try to read undefined attributes from standard headers
              if result.name.nil? && headers.contains(:content_disposition) && headers[:content_disposition][0] =~ /filename=([^;]+)/
                name = $1
                name = name[1..-2] if name =~ /^".*"$/
                result.name = name.empty? ? nil : name
              end
              if result.content_type.nil? && headers.contains(:content_type)
                result.content_type = headers[:content_type][0]
              end
              if result.size.nil? && headers.contains(:content_length)
                result.size = headers[:content_length][0].to_i
              end
              result.data = data
            else
              result = http_response.body
              result = result.read if result.kind_of?(IO)
              result.force_encoding(@@DEFAULT_ENCODING)
            end
          end
          return result
        end

        # Implementation of {UU::OS::CMD::CommandHandler}.
        # @private
        def handle_error(http_response, options) # :nodoc:
          error_code = http_response.status
          error_body = http_response.body
          error_body = error_body.read if error_body.kind_of?IO
          # Try to handle potential error by wrapping service
          if @service.respond_to?:process_error
            error = @service.process_error(error_code, error_body, http_response.headers)
            return error if error.kind_of?Exception
          end
          # Use default error handling
          if error_code == 401
            return UU::OS::Security::AuthenticationException.new('Invalid credentials.')
          elsif error_code == 403
            return UU::OS::Security::AuthorizationException.new(error_body)
          elsif error_code == 404
            if (error_body =~ /^{.*}$/)
              # Probably contains JSON body therefore it is likely exception thrown by application.
              return UU::OS::REST::ClientException.new(error_body)
            else
              # Generic 404 exception provided by target server. Translate it to something more readable.
              return UU::OS::REST::ClientException.new("Service at #{http_response.request.url} is not available. Either it is offline or configuration in uuHome points to wrong URL.")
            end
          elsif error_code == 429
            return UU::OS::QoS::QoSLimitException.new(error_body)
          elsif error_code == 503
            if (error_body =~ /^{.*}$/)
              # Probably contains JSON body therefore it is likely exception thrown by application.
              return UU::OS::REST::ServerException.new(error_body)
            else
              # Generic 503 exception provided by target server. Translate it to something more readable.
              return UU::OS::REST::ServerException.new("Service at #{http_response.request.url} is not available. It is probably offline.")
            end
          elsif error_code < 500
            exception = nil
            begin
              exception = JSON.parse(error_body, :symbolize_names => true) if error_body =~ /^{.*}$/
            rescue Exception => e
              # unparsable exception, will be raised as common client exception
            end
            return UU::OS::REST::ClientException.new(error_body) if exception.nil?
            if exception[:exceptionClass] =~ /MainEntity.*RTException|TerritorySwitchServiceRTException/
              return UU::OS::REST::MainEntityException.new(exception)
            elsif ['UU.OS/E05300.M00', 'UU.OS/E05207.M00', 'UU.OS/E05205.M00', 'UU.OS/E05205.M02', 'UU.OS/E05205.M04'].include?exception[:code]
              return UU::OS::REST::MainEntityException.new(exception)
            else
              return UU::OS::REST::ClientException.new(exception)
            end
          else
            return UU::OS::REST::ServerException.new(error_body)
          end
        end

        # Moves attributes of nested structures to first level.
        # @private
        def flatten_nested_structures(input, output, parent_name = nil) # :nodoc:
          input.each do |key, value|
            if parent_name.nil? || parent_name.empty?
              name = key.to_s
            else
              name = "#{parent_name}[#{key}]"
            end
            if value.kind_of?Hash
              flatten_nested_structures(value, output, name)
            else
              output[name] = value
            end
          end
        end

        # ==================================================================
        # This private method is not needed in new implementation, however is
        # used from Property, therefore we have to keep it for now.
        # ==================================================================

        # Check for errors and process response.
        #
        # @param [Array<Numeric, String, Hash>] result Result of raw request method
        #
        # @return [String] Response body (or throws exception)
        #
        # @private
        def process_result(result) # :nodoc:
          # Service was successfully invoked, no error occurred
          if result[0] < 300
            return result[1]
          end
          if result[1].kind_of?IO
            error_body = result[1].read
            result[1].close
          else
            error_body = result[1]
          end
          # Try to handle potential error by wrapping service
          if @service.respond_to?:process_error
            error = @service.process_error(result[0], error_body, result[2])
            raise error if error.kind_of?Exception
          end
          # Use default error handling
          if result[0] == 401
            raise UU::OS::Security::AuthenticationException.new('Invalid credentials.')
          elsif result[0] == 403
            raise UU::OS::Security::AuthorizationException.new(error_body)
          elsif result[0] == 404
            if (error_body =~ /^{.*}$/)
              # Probably contains JSON body therefore it is likely exception thrown by application.
              raise UU::OS::REST::ClientException.new(error_body)
            else
              # Generic 404 exception provided by target server. Translate it to something more readable.
              url = "#{@client.unwrap.instance_variable_get(:@base_url)}/#{@base_path}"
              raise UU::OS::REST::ClientException.new("Service at #{url} is not available. Either it is offline or configuration in uuHome points to wrong URL.")
            end
          elsif result[0] == 429
            raise UU::OS::QoS::QoSLimitException.new(error_body)
          elsif result[0] == 503
            if (error_body =~ /^{.*}$/)
              # Probably contains JSON body therefore it is likely exception thrown by application.
              raise UU::OS::REST::ServerException.new(error_body)
            else
              # Generic 503 exception provided by target server. Translate it to something more readable.
              url = "#{@client.unwrap.instance_variable_get(:@base_url)}/#{@base_path}"
              raise UU::OS::REST::ServerException.new("Service at #{url} is not available. It is probably offline.")
            end
          elsif result[0] < 500
            exception = nil
            begin
              exception = JSON.parse(error_body, :symbolize_names => true) if error_body =~ /^{.*}$/
            rescue Exception => e
              # unparsable exception, will be raised as common client exception
            end
            raise UU::OS::REST::ClientException.new(error_body) if exception.nil?
            if exception[:exceptionClass] =~ /MainEntity.*RTException|TerritorySwitchServiceRTException/
              raise UU::OS::REST::MainEntityException.new(exception)
            elsif ['UU.OS/E05300.M00', 'UU.OS/E05207.M00', 'UU.OS/E05205.M00', 'UU.OS/E05205.M02', 'UU.OS/E05205.M04'].include?exception[:code]
              raise UU::OS::REST::MainEntityException.new(exception)
            else
              raise UU::OS::REST::ClientException.new(exception)
            end
          else
            raise UU::OS::REST::ServerException.new(error_body)
          end
        end

      end

      # Binary stream temporary file wrapper.
      # @private
      class TempfileIO < IO # :nodoc:

        # Create new instance of stream.
        #
        # @param [Tempfile, IO] data Data to be wrapped
        def initialize(data)
          if data.kind_of?(IO) && !data.kind_of?(Tempfile)
            begin
              temp = Tempfile.new('lbr')
              temp.binmode
              while (!(chunk = data.read(65535)).nil?)
                temp << chunk
              end
              temp.flush
            ensure
              temp.close if !temp.nil?
              data.close
            end
            data = temp
          end
          super(data.open.fileno, 'rb')
          @tempfile = data
          # On POSIX systems it is possible to call unlink of temporary file before reading and closing it.
          # (see http://www.ruby-doc.org/stdlib-1.9.3/libdoc/tempfile/rdoc/Tempfile.html#label-Unlink+after+creation)
          # Suppress warning which occurs on JRuby < 1.7.17
          orig_verbose = $VERBOSE
          begin
            $VERBOSE = nil
            data.unlink
          ensure
            $VERBOSE = orig_verbose
          end
        end

        # Gets size of stream
        def size
          @tempfile.size
        end

        # Closes stream and disposes temporary file.
        def close
          begin
            super
          ensure
            dispose_tempfile
          end
        end

        # Closes stream and disposes temporary file.
        def close_read
          begin
            super
          ensure
            dispose_tempfile
          end
        end

        private

        # Disposes wrapped temporary file.
        def dispose_tempfile
          return if @tempfile.nil?
          # We do not call "close!" because it internally
          # calls "close" which may file and "unlink" is
          # never called in that case.
          begin
            @tempfile.close if !@tempfile.closed?
          rescue
            # Close may fail if file was physically
            # deleted before closing it - no problem
          end
          @tempfile.unlink
          @tempfile = nil
        end

      end

    end
  end
end
